/****************************************************************************

    Copyright (c) 1994 by SimSoft

    PROGRAM: testapp.c

    PURPOSE: Test program to exercise the functionality of ImgLib

****************************************************************************/

#include <windows.h>
#include <commdlg.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include "tstapp.h"
#include "imglib.h"

#ifdef WIN32
#define huge
#endif
typedef char huge * HPSTR;

HWND              hwndMain;
HINSTANCE         hInst;
char              szBuffer[100];
HINSTANCE         hWinGInst;
HACCEL            hAccel;

BOOL InitApplication (HANDLE);
BOOL InitInstance (HANDLE, int);
long WINAPI MainWndProc (HWND, UINT, WPARAM, LPARAM);
BOOL GetOpenFile (LPSTR);
BOOL GetSaveFile (LPSTR);
LPVOID CreateDisplayDIB (HWND, LPVOID, HPALETTE *);

/****************************************************************************

    FUNCTION: WinMain(HANDLE, HANDLE, LPSTR, int)

    PURPOSE: calls initialization function, processes message loop

****************************************************************************/

int PASCAL WinMain (HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{ 
MSG               msg;

  hInst = hInstance;

  if (!hPrevInstance)
  {
    if (!InitApplication (hInstance))
      return (FALSE);
  }
  
  if (!InitInstance (hInstance, nCmdShow))
    return (FALSE);
  
  while (GetMessage (&msg, NULL, 0, 0xffff))
  {
    if (!TranslateAccelerator (hwndMain, hAccel, &msg))
    {
      TranslateMessage (&msg);
      DispatchMessage (&msg);
    }
  }

  return (0);
}


/****************************************************************************

    FUNCTION: InitApplication(HANDLE)

    PURPOSE: Initializes window data and registers window class

****************************************************************************/

BOOL InitApplication (HANDLE hInstance)
{
WNDCLASS          wc;

  wc.style = 0;
  wc.lpfnWndProc = MainWndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon (NULL, IDI_APPLICATION);
  wc.hCursor = LoadCursor (NULL, IDC_ARROW);
  wc.hbrBackground = GetStockObject (WHITE_BRUSH);
  wc.lpszMenuName = "tstappmenu";
  wc.lpszClassName = "Test App";
  
  return (RegisterClass (&wc));
}


/****************************************************************************

    FUNCTION:  InitInstance(HANDLE, int)

    PURPOSE:  Saves instance handle and creates main window

****************************************************************************/

BOOL InitInstance (HANDLE hInstance, int nCmdShow)
{
UINT              uErrMode;
HMENU             hMenu;

  hAccel = LoadAccelerators (hInstance, "tstappaccel");

  if (!hAccel)
    return (FALSE);

  /*
   * Just use default size and location.  It's easier to size an already
   * created window for the specific client area size.
   */
   
  hwndMain = CreateWindow ("Test App", "Test App",
        WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
        NULL, NULL, hInstance, NULL);

  if (!hwndMain)
    return (FALSE);

  ShowWindow (hwndMain, nCmdShow);
  UpdateWindow (hwndMain);

  uErrMode = SetErrorMode (SEM_NOOPENFILEERRORBOX);
#ifdef WIN32
  hWinGInst = LoadLibrary ("wing32.dll");
#else
  hWinGInst = LoadLibrary ("wing.dll");
  
  if (hWinGInst < HINSTANCE_ERROR)
    hWinGInst = 0;
#endif
  SetErrorMode (uErrMode);
  
  if (!hWinGInst)
  {
    hMenu = GetMenu (hwndMain);
    EnableMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_WING, MF_BYCOMMAND | MF_GRAYED);
  }
  
  return (TRUE);
}

/****************************************************************************

    FUNCTION: MainWndProc(HWND, UINT, WPARAM, LPARAM)

    PURPOSE:  Processes messages

    MESSAGES:

  WM_COMMAND    - application menu (About dialog box)
  WM_DESTROY    - destroy window

****************************************************************************/

long WINAPI MainWndProc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
int               iCount, iXOverhead, iYOverhead;
static LPVOID     lpDIB1, lpDIB2, lpDIBDisplay;
static HPALETTE   hPal;
HPALETTE          hPalOld;
HDC               hDC;
LPBITMAPINFO      pBitmapInfo;
PAINTSTRUCT       PaintStruct;
static int        iUpdateCount;
char              szFileName[260];
LPVOID            lpDIBNew;
HMENU             hMenu;
HPSTR             hpBits;

  switch (message)
  {
    case WM_ACTIVATE:
      
      /*
       * Check if app. is being de-activated.  If not, fall through to update
       * the palette just as in WM_QUERYNEWPALETTE processing.
       */

      if (!wParam)
        break;

    case WM_QUERYNEWPALETTE:

    /*
     * If palette realization causes a palette change,
     * we need to do a full redraw.
     */

    if (hPal && wParam != (UINT)hWnd)
    {
      hDC = GetDC (hWnd);
      hPalOld = SelectPalette (hDC, hPal, FALSE);
      iCount = RealizePalette (hDC);
      SelectPalette (hDC, hPalOld, 0);
      ReleaseDC (hWnd, hDC);

      if (iCount)
      {
        iUpdateCount = 0;
        InvalidateRect (hWnd, (LPRECT) (NULL), 0);
        return TRUE;
      }
      else
        return FALSE;
    }
    break;

    case WM_PALETTECHANGED:

      if (hPal && wParam != (UINT)hWnd)
      {
        hDC = GetDC (hWnd);
        hPalOld = SelectPalette (hDC, hPal, FALSE);
        iCount = RealizePalette (hDC);

        if (iCount)
        {
          UpdateColors (hDC);
          iUpdateCount++;
        }

        SelectPalette (hDC, hPalOld, FALSE);
        ReleaseDC (hWnd, hDC);
      }

      break;

    case WM_PAINT:
      
      /*
       * If we have updated more than once, the image quality may have
       * degraded too much.  Force a full repaint.
       */

      if (iUpdateCount > 1)
      {
        BeginPaint (hWnd, &PaintStruct);
        EndPaint (hWnd, &PaintStruct);
        iUpdateCount = 0;
        InvalidateRect (hWnd, (LPRECT) (NULL), 0);
        break;
      }

      hDC = BeginPaint (hWnd, &PaintStruct);

      if (lpDIBDisplay)
      {
        if (hPal)
        {
          hPalOld = SelectPalette (hDC, hPal, FALSE);
          RealizePalette (hDC);
        }

        pBitmapInfo = (LPBITMAPINFO) lpDIBDisplay;
        hpBits = (HPSTR)&pBitmapInfo->bmiColors[pBitmapInfo->bmiHeader.biClrUsed];
        SetDIBitsToDevice (hDC, 0, 0, (int)pBitmapInfo->bmiHeader.biWidth,
              (int)pBitmapInfo->bmiHeader.biHeight, 0, 0, 0,
              (int)pBitmapInfo->bmiHeader.biHeight, hpBits,
              pBitmapInfo, DIB_RGB_COLORS);

        if (hPal)
        {
          SelectPalette (hDC, hPalOld, FALSE);
        }
      }

      EndPaint (hWnd, &PaintStruct);
      break;

    case WM_COMMAND:

#ifdef WIN32
      switch (LOWORD (wParam))
#else
      switch (wParam)
#endif
      {
        case ID_FILE_SAVE:

          if (lpDIB1)
          {
            if (GetSaveFile (szFileName) == FALSE)
              break;

            if (WriteDIBToFile (lpDIB1, szFileName) == FALSE)
            {
              switch (GetLastImgLibError ())
              {
                case ERROR_WRITE_ACCESS_DENIED:

                  sprintf (szBuffer, "Error writing to file \"%s\"", szFileName);
                  break;

                case ERROR_INVALID_POINTER:

                  sprintf (szBuffer, "Error accessing one of the pointer parameters", szFileName);
                  break;
              }

              MessageBox (NULL, szBuffer, "Save Error", MB_TASKMODAL | MB_ICONSTOP | MB_OK);
            }
          }
          break;

        case ID_FILE_OPEN:
        
          if (GetOpenFile (szFileName) == FALSE)
            break;

          lpDIBNew = ReadFileIntoDIB (szFileName);
          
          if (lpDIBNew)
          {
            if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
              DIBFree (lpDIBDisplay);

            if (lpDIB2)
              DIBFree (lpDIB2);

            lpDIB2 = lpDIB1;
            lpDIB1 = lpDIBNew;

            if (hPal)
            {
              DeleteObject (hPal);
              hPal = NULL;
            }
            
            lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
            
            if (lpDIBDisplay)
            {
              iXOverhead = GetSystemMetrics (SM_CXFRAME) * 2;
              iYOverhead = GetSystemMetrics (SM_CYFRAME) * 2 + GetSystemMetrics (SM_CYCAPTION) + GetSystemMetrics (SM_CYMENU) - GetSystemMetrics (SM_CYBORDER);
              pBitmapInfo = (LPBITMAPINFO) lpDIBDisplay;
              SetWindowPos (hWnd, NULL, 0, 0,
                  (int)pBitmapInfo->bmiHeader.biWidth + iXOverhead,
                  (int)pBitmapInfo->bmiHeader.biHeight + iYOverhead,
                  SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
              InvalidateRect (hWnd, (LPRECT) (NULL), TRUE);
            }
          }
          break;

        case ID_FILE_EXIT:

          DestroyWindow (hWnd);
          break;

        case ID_EDIT_UNDO:

          if (lpDIB1 && lpDIB2)
          {
            if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
              DIBFree (lpDIBDisplay);

            lpDIBNew = lpDIB2;
            lpDIB2 = lpDIB1;
            lpDIB1 = lpDIBNew;

            if (hPal)
            {
              DeleteObject (hPal);
              hPal = NULL;
            }
            
            lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
            InvalidateRect (hWnd, (LPRECT) (NULL), FALSE);
          }
          break;

        case ID_PROCESS_BRIGHTEN:
        case ID_PROCESS_DARKEN:

          if (lpDIB1)
          {
#ifdef WIN32
            if (LOWORD (wParam) == ID_PROCESS_BRIGHTEN)
#else
            if (wParam == ID_PROCESS_BRIGHTEN)
#endif
              lpDIBNew = BrightenDIB (lpDIB1, 110);
            else
              lpDIBNew = BrightenDIB (lpDIB1, 90);

            if (lpDIBNew)
            {
              if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
                DIBFree (lpDIBDisplay);

              if (lpDIB2)
                DIBFree (lpDIB2);

              lpDIB2 = lpDIB1;
              
              if (hPal)
                DeleteObject (hPal);

              lpDIB1 = lpDIBNew;
              lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
              InvalidateRect (hWnd, (LPRECT) (NULL), FALSE);
            }
          }
          break;

        case ID_PROCESS_SMOOTH:

          if (lpDIB1)
          {
            lpDIBNew = SmoothDIB (lpDIB1, 1);

            if (lpDIBNew)
            {
              if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
                DIBFree (lpDIBDisplay);

              if (lpDIB2)
                DIBFree (lpDIB2);

              lpDIB2 = lpDIB1;
              
              if (hPal)
                DeleteObject (hPal);

              lpDIB1 = lpDIBNew;
              lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
              InvalidateRect (hWnd, (LPRECT) (NULL), FALSE);
            }
          }
          break;

        case ID_PROCESS_HALFTONE:

          if (lpDIB1)
          {
            lpDIBNew = HalftoneDIB (lpDIB1);

            if (lpDIBNew)
            {
              if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
                DIBFree (lpDIBDisplay);

              if (lpDIB2)
                DIBFree (lpDIB2);

              lpDIB2 = lpDIB1;
              
              if (hPal)
                DeleteObject (hPal);

              lpDIB1 = lpDIBNew;
              lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
              InvalidateRect (hWnd, (LPRECT) (NULL), TRUE);
            }
          }
          break;

        case ID_PROCESS_GRAY:

          if (lpDIB1)
          {
            lpDIBNew = GrayDIB (lpDIB1);

            if (lpDIBNew)
            {
              if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
                DIBFree (lpDIBDisplay);

              if (lpDIB2)
                DIBFree (lpDIB2);

              lpDIB2 = lpDIB1;
              
              if (hPal)
                DeleteObject (hPal);

              lpDIB1 = lpDIBNew;
              lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
              InvalidateRect (hWnd, (LPRECT) (NULL), FALSE);
            }
          }
          break;

        case ID_OPTIONS_TC_DISPLAY_GDI:

          hMenu = GetMenu (hWnd);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_GDI, MF_BYCOMMAND | MF_CHECKED);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_WING, MF_BYCOMMAND | MF_UNCHECKED);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_IMGLIB, MF_BYCOMMAND | MF_UNCHECKED);

          if (lpDIB1)
          {
            if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
              DIBFree (lpDIBDisplay);

            lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
            InvalidateRect (hWnd, (LPRECT) (NULL), FALSE);
          }

          break;

        case ID_OPTIONS_TC_DISPLAY_WING:

          hMenu = GetMenu (hWnd);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_GDI, MF_BYCOMMAND | MF_UNCHECKED);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_WING, MF_BYCOMMAND | MF_CHECKED);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_IMGLIB, MF_BYCOMMAND | MF_UNCHECKED);

          if (lpDIB1)
          {
            if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
              DIBFree (lpDIBDisplay);

            lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
            InvalidateRect (hWnd, (LPRECT) (NULL), FALSE);
          }

          break;

        case ID_OPTIONS_TC_DISPLAY_IMGLIB:

          hMenu = GetMenu (hWnd);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_GDI, MF_BYCOMMAND | MF_UNCHECKED);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_WING, MF_BYCOMMAND | MF_UNCHECKED);
          CheckMenuItem (hMenu, ID_OPTIONS_TC_DISPLAY_IMGLIB, MF_BYCOMMAND | MF_CHECKED);

          if (lpDIB1)
          {
            if (lpDIBDisplay && lpDIBDisplay != lpDIB1)
              DIBFree (lpDIBDisplay);

            lpDIBDisplay = CreateDisplayDIB (hWnd, lpDIB1, &hPal);
            InvalidateRect (hWnd, (LPRECT) (NULL), FALSE);
          }

          break;

        default:
          return (DefWindowProc (hWnd, message, wParam, lParam));
      }
      break;

    case WM_DESTROY:

      if (lpDIB1)
        DIBFree (lpDIB1);

      if (lpDIB2)
        DIBFree (lpDIB2);

      if (hPal)
      {
        DeleteObject (hPal);
        hPal = NULL;
      }

      if (hWinGInst)
        FreeLibrary (hWinGInst);

      PostQuitMessage(0);
      break;

    default:
      return (DefWindowProc (hWnd, message, wParam, lParam));
  }
  return (0);
}

/*************************************************************************

  Function:  GetOpenFile (LPSTR. WORD)

   Purpose:  Prompts user for a filename through the use of a Windows
             FileOpen common dialog box.

   Returns:  TRUE if a filename is selected.
             FALSE if no filename is selected.

  Comments:  Filename is put into the string passed to the routine.
             If a filename is not selected, NULL is returned.

*************************************************************************/

BOOL GetOpenFile (LPSTR szFileName)
{
   OPENFILENAME   of;
   DWORD          flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_NOCHANGEDIR | OFN_PATHMUSTEXIST;
   static char    szTitle[30];         // Dialog Box title
   static char    szTemplate[20];      // Dialog Box template
   static char    szFile[256];         // File name
   static char    szFileTitle[256];    // Title
   static char    szDrive[5];          // Drive
   static char    szDir[256];          // Directory
   char szFilter[] =                   // Filter
      "Pictures\0*.bmp;*.jpg;*.gif;*.tga;*.pcd\0";
   char          *pFile;


   // Initialize the OPENFILENAME members

   szFile[0] = '\0';

   of.lStructSize       = sizeof (OPENFILENAME);
   of.hwndOwner         = GetFocus ();
   of.hInstance         = hInst;
   of.lpstrFilter       = szFilter;
   of.lpstrCustomFilter = NULL;
   of.nMaxCustFilter    = 0L;
   of.nFilterIndex      = 1L;
   of.lpstrFile         = szFile;
   of.nMaxFile          = sizeof (szFile);
   of.lpstrFileTitle    = szFileTitle;
   of.nMaxFileTitle     = sizeof (szFileTitle);
   of.lpstrInitialDir   = szDir;
   of.lpstrTitle        = "Open a new test bitmap file";
   of.Flags             = flags;
   of.nFileOffset       = 0;
   of.nFileExtension    = 0;
   of.lpstrDefExt       = NULL;
   of.lCustData         = 0;
   of.lpfnHook          = NULL;
   of.lpTemplateName    = NULL;


   // Call the GetOpenFilename function

  if (GetOpenFileName (&of))
  {
    lstrcpy (szFileName, of.lpstrFile);
    lstrcpy (szDir, of.lpstrFile);
    pFile = strrchr (szDir, '\\');

    if (pFile)
      *pFile = 0;

    return TRUE;
  }
  else
    return FALSE;
}

/*************************************************************************

  Function:  GetSaveFile (LPSTR. WORD)

   Purpose:  Prompts user for a filename through the use of a Windows
             FileOpen common dialog box.

   Returns:  TRUE if a filename is selected.
             FALSE if no filename is selected.

  Comments:  Filename is put into the string passed to the routine.
             If a filename is not selected, NULL is returned.

*************************************************************************/

BOOL GetSaveFile (LPSTR szFileName)
{
   OPENFILENAME   of;
   DWORD          flags = OFN_HIDEREADONLY | OFN_NOCHANGEDIR | OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT;
   static char    szTitle[30];         // Dialog Box title
   static char    szTemplate[20];      // Dialog Box template
   static char    szFile[256];         // File name
   static char    szFileTitle[256];    // Title
   static char    szDrive[5];          // Drive
   static char    szDir[256];          // Directory
   char szFilter[] =                   // Filter
      "Windows Pictures\0*.bmp\0";
   char          *pFile;


   // Initialize the OPENFILENAME members

   szFile[0] = '\0';

   of.lStructSize       = sizeof (OPENFILENAME);
   of.hwndOwner         = GetFocus ();
   of.hInstance         = hInst;
   of.lpstrFilter       = szFilter;
   of.lpstrCustomFilter = NULL;
   of.nMaxCustFilter    = 0L;
   of.nFilterIndex      = 1L;
   of.lpstrFile         = szFile;
   of.nMaxFile          = sizeof (szFile);
   of.lpstrFileTitle    = szFileTitle;
   of.nMaxFileTitle     = sizeof (szFileTitle);
   of.lpstrInitialDir   = szDir;
   of.lpstrTitle        = "Save the image in a bitmap file";
   of.Flags             = flags;
   of.nFileOffset       = 0;
   of.nFileExtension    = 0;
   of.lpstrDefExt       = "*.bmp";
   of.lCustData         = 0;
   of.lpfnHook          = NULL;
   of.lpTemplateName    = NULL;


   // Call the GetOpenFilename function

  if (GetSaveFileName (&of))
  {
    lstrcpy (szFileName, of.lpstrFile);
    lstrcpy (szDir, of.lpstrFile);
    pFile = strrchr (szDir, '\\');

    if (pFile)
      *pFile = 0;

    return TRUE;
  }
  else
    return FALSE;
}

/*************************************************************************

  Function:  CreateDisplayDIB()

   Purpose:  Creates a DIB suitable for the display based on user-selected options.

   Returns:  a new DIB pointer

*************************************************************************/

LPVOID CreateDisplayDIB (HWND hwnd, LPVOID lpDIB, HPALETTE *phPal)
{
HMENU             hMenu;
LPBITMAPINFO      pBitmapInfo;
UINT              uState;
LPVOID            lpDIBDisplay;
long              lColors;
int               iPixelBits, iPlanes;
HDC               hDC;

  hMenu = GetMenu (hwnd);
  uState = GetMenuState (hMenu, ID_OPTIONS_TC_DISPLAY_GDI, MF_BYCOMMAND);
  pBitmapInfo = (LPBITMAPINFO) lpDIB;

  if (uState & MF_CHECKED || pBitmapInfo->bmiHeader.biClrUsed)
  {
    lpDIBDisplay = lpDIB;
    
    if (pBitmapInfo->bmiHeader.biClrUsed)
      *phPal = CreateDIBPalette (lpDIBDisplay);
    else
      *phPal = NULL;
  }
  else
  {
    uState = GetMenuState (hMenu, ID_OPTIONS_TC_DISPLAY_WING, MF_BYCOMMAND);

    if (uState & MF_CHECKED && hWinGInst)
    {
      lpDIBDisplay = WinGHalftoneDIB (lpDIB, phPal);
    }
    else
    {
      hDC = GetDC (hwnd);
      iPixelBits = GetDeviceCaps (hDC, BITSPIXEL);
      iPlanes = GetDeviceCaps (hDC, PLANES);
      lColors = 1 << iPixelBits * iPlanes;
      ReleaseDC (hwnd, hDC);
      lpDIBDisplay = ReduceDIB (lpDIB, lColors, TRUE);
      *phPal = CreateDIBPalette (lpDIBDisplay);
    }
  }

  return (lpDIBDisplay);
}
